#
#   2025 Fabian Jankowski
#   Normal - lognormal mixture model.
#

import logging

import numpy as np
import pymc as pm

from fitpdf.models.model import Model


class NL(Model):
    name = "Normal - lognormal"

    def __init__(self):
        """
        Model distribution.
        """

        self.__log = logging.getLogger("fitpdf.models.nl")

        self.ncomp = 2

    def __repr__(self):
        """
        Representation of the object.
        """

        info_dict = {"bla": "XXX"}

        info_str = "{0}".format(info_dict)

        return info_str

    def __str__(self):
        """
        String representation of the object.
        """

        info_str = "{0}: {1}".format(self.name, repr(self))

        return info_str

    def get_model(self, t_data, t_offp):
        """
        Construct a mixture model.

        Parameters
        ----------
        t_data: ~np.array of float
            The input data to be fit.
        t_offp: ~np.array of float
            The off-pulse data.

        Returns
        -------
        model: ~pm.Model
            A mixture model.
        """

        data = t_data.copy()
        offp = t_offp.copy()

        # on-pulse mean and std
        onp_mean = np.mean(data)
        onp_std = np.std(data)
        print(f"On-pulse mean: {onp_mean:.5f}")
        print(f"On-pulse std: {onp_std:.5f}")

        # off-pulse mean and std
        offp_mean = np.mean(offp)
        offp_std = np.std(offp)
        print(f"Off-pulse mean: {offp_mean:.5f}")
        print(f"Off-pulse std: {offp_std:.5f}")

        coords = {"component": np.arange(2), "obs_id": np.arange(len(data))}

        with pm.Model(coords=coords) as model:
            x = pm.Data("x", data, dims="obs_id")

            # mixture weights
            w = pm.Dirichlet("w", a=np.array([0.3, 0.7]), dims="component")

            # priors
            mu = pm.Normal(
                "mu",
                mu=np.array([offp_mean, np.log(onp_mean)]),
                sigma=np.array([0.1, np.log(1.75)]),
                dims="component",
            )
            sigma = pm.HalfNormal(
                "sigma",
                sigma=np.array([offp_std, 1.0]),
                dims="component",
            )

            # 1) normal distribution for nulling
            # mu = location, sigma = scale
            norm = pm.Normal.dist(mu=mu[0], sigma=sigma[0])

            # 2) lognormal distribution for pulses
            # mu = log of location, sigma = log of scale
            lognorm = pm.Lognormal.dist(mu=mu[1], sigma=sigma[1])

            components = [norm, lognorm]

            pm.Mixture("obs", w=w, comp_dists=components, observed=x, dims="obs_id")

        return model

    def get_analytic_pdf(self, x, w, mu, sigma, icomp):
        """
        Get the analytic PDF.

        Returns
        -------
        pdf: ~np.array of float
            The model PDF evaluated at the `x` values.
        """

        if icomp == 0:
            dist = pm.Normal.dist(mu=mu, sigma=sigma)
        elif icomp == 1:
            dist = pm.Lognormal.dist(mu=mu, sigma=sigma)
        else:
            raise NotImplementedError(f"Component not implemented: {icomp}")

        pdf = w[icomp] * pm.logp(dist, x).exp()

        return pdf.eval()

    def get_mode(self, mu, sigma, icomp):
        """
        Compute the mode of the model component.

        Returns
        -------
        mode: ~np.array of float
            The mode of the model component.
        """

        if icomp == 0:
            mode = mu
        elif icomp == 1:
            mode = np.exp(mu - np.power(sigma, 2))
        else:
            raise NotImplementedError(f"Component not implemented: {icomp}")

        return mode



#
#   2025 Fabian Jankowski
#   Simulate distributions.
#

import argparse
import logging
import os
import signal
import sys

import numpy as np
import pandas as pd
import pymc as pm

if "DISPLAY" not in os.environ:
    # set a rendering backend that does not require an X server
    matplotlib.use("Agg")
import matplotlib.pyplot as plt

from fitpdf.general_helpers import (
    configure_logging,
    customise_matplotlib_format,
    signal_handler,
)


def parse_args():
    """
    Parse the commandline arguments.

    Returns
    -------
    args: populated namespace
        The commandline arguments.
    """

    parser = argparse.ArgumentParser(
        description="Simulate distributions.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--nsamp",
        dest="nsamp",
        type=int,
        metavar=("value"),
        default=10000,
        help="Number of random samples to draw from the simulated distribution.",
    )

    # options that affect the output formatting
    output = parser.add_argument_group(title="Output formatting")

    output.add_argument(
        "-o",
        "--output",
        dest="output",
        action="store_true",
        default=False,
        help="Output plots to file rather than to screen.",
    )

    args = parser.parse_args()

    return args


def check_args(args):
    """
    Sanity check the commandline arguments.

    Parameters
    ----------
    args: populated namespace
        The commandline arguments.
    """

    log = logging.getLogger("fitpdf.simulate_dist")

    # nsamp
    if not args.nsamp > 100:
        log.error(f"Number of samples is invalid: {args.nsamp}")
        sys.exit(1)


#
# MAIN
#


def main():
    # start signal handler
    signal.signal(signal.SIGINT, signal_handler)

    # set up logging
    configure_logging()
    log = logging.getLogger("fitpdf.simulate_dist")

    # handle command line arguments
    args = parse_args()

    # sanity check command line arguments
    check_args(args)

    # tweak the matplotlib output formatting
    customise_matplotlib_format()

    params = {
        "dpi": 300,
        "nsamp": args.nsamp,
        "output": args.output,
    }

    weights = [0.3, 0.3, 0.7]
    mu = [0.0, 0.3, np.log(1.75)]
    sigma = [2.0, 1.0, 1.0]

    # off-pulse
    foff = pm.Normal.dist(mu=mu[0], sigma=sigma[0])

    foff_samples = pm.draw(foff, draws=params["nsamp"])

    # on-pulse
    fon = pm.Mixture.dist(
        w=weights,
        comp_dists=[
            pm.Normal.dist(mu=mu[0], sigma=sigma[0]),
            pm.Normal.dist(mu=mu[1], sigma=sigma[1]),
            pm.LogNormal.dist(mu=mu[2], sigma=sigma[2]),
        ],
    )

    fon_samples = pm.draw(fon, draws=params["nsamp"])

    # write to disk
    _temp = {
        "rotation": np.arange(params["nsamp"]),
        "zapped": np.zeros(params["nsamp"]).astype(int),
        "fluence_on": fon_samples,
        "nbin_on": 64,
        "fluence_off": foff_samples,
        "nbin_off": 64,
        "fluence_off_same": foff_samples,
        "nbin_off_same": 64,
    }
    _df = pd.DataFrame(_temp)
    _df.to_csv("simulated_fluences.csv", index=False)

    fig = plt.figure()
    ax = fig.add_subplot()

    ax.hist(
        foff_samples,
        bins="auto",
        color="dimgrey",
        density=True,
        histtype="stepfilled",
        zorder=3,
        alpha=0.3,
    )

    ax.hist(
        fon_samples, bins="auto", color="black", density=True, histtype="step", zorder=5
    )

    ax.set_xlabel(r"$F \: / \: \left< F_\mathrm{on} \right>$")
    ax.set_ylabel("PDF")
    ax.set_yscale("log")

    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "simulated_pdf.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)

    plt.show()

    log.info("All done.")


if __name__ == "__main__":
    main()



#
#   2025 Fabian Jankowski
#   Lognormal - lognormal mixture model.
#

import logging

import numpy as np
import pymc as pm

from fitpdf.models.model import Model


class LL(Model):
    name = "Lognormal - lognormal"

    def __init__(self):
        """
        Model distribution.
        """

        self.__log = logging.getLogger("fitpdf.models.ll")

        self.ncomp = 2

    def __repr__(self):
        """
        Representation of the object.
        """

        info_dict = {"bla": "XXX"}

        info_str = "{0}".format(info_dict)

        return info_str

    def __str__(self):
        """
        String representation of the object.
        """

        info_str = "{0}: {1}".format(self.name, repr(self))

        return info_str

    def get_model(self, t_data, t_offp):
        """
        Construct a mixture model.

        Parameters
        ----------
        t_data: ~np.array of float
            The input data to be fit.
        t_offp: ~np.array of float
            The off-pulse data.

        Returns
        -------
        model: ~pm.Model
            A mixture model.
        """

        data = t_data.copy()
        offp = t_offp.copy()

        with pm.Model() as model:
            # mixture weights
            w = pm.Dirichlet("w", a=np.array([1, 1]))

            # priors
            mu = pm.Normal("mu", mu=np.array([0, 1]), sigma=1)
            sigma = pm.HalfNormal("sigma", sigma=np.array([1, 1]))

            # 1) lognormal distribution
            lognorm1 = pm.Lognormal.dist(mu=mu[0], sigma=sigma[0])

            # 2) lognormal distribution
            lognorm2 = pm.Lognormal.dist(mu=mu[1], sigma=sigma[1])

            components = [lognorm1, lognorm2]

            pm.Mixture("obs", w=w, comp_dists=components, observed=data)

        return model

    def get_analytic_pdf(self, x, w, mu, sigma, icomp):
        """
        Get the analytic PDF.

        Returns
        -------
        pdf: ~np.array of float
            The model PDF evaluated at the `x` values.
        """

        if icomp in [0, 1]:
            dist = pm.Lognormal.dist(mu=mu, sigma=sigma)
        else:
            raise NotImplementedError(f"Component not implemented: {icomp}")

        pdf = w[icomp] * pm.logp(dist, x).exp()

        return pdf.eval()

    def get_mode(self, mu, sigma, icomp):
        """
        Compute the mode of the model component.

        Returns
        -------
        mode: ~np.array of float
            The mode of the model component.
        """

        if icomp in [0, 1]:
            mode = np.exp(mu - np.power(sigma, 2))
        else:
            raise NotImplementedError(f"Component not implemented: {icomp}")

        return mode



import os.path
from setuptools import setup


def get_version():
    version_file = os.path.join(os.path.dirname(__file__), "fitpdf", "version.py")

    with open(version_file, "r") as f:
        raw = f.read()

    items = {}
    exec(raw, None, items)

    return items["__version__"]


setup(
    version=get_version(),
)



#
#   2025 Fabian Jankowski
#   Fit measured distribution data.
#

import argparse
import logging
import os
import signal
import sys

import arviz as az

# switch between interactive and non-interactive mode
import matplotlib

if "DISPLAY" not in os.environ:
    # set a rendering backend that does not require an X server
    matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pymc as pm

from fitpdf.general_helpers import (
    configure_logging,
    customise_matplotlib_format,
    signal_handler,
)
from spanalysis.apps.plot_dist import plot_pe_dist
import fitpdf.models as fmodels
from fitpdf.plotting import plot_chains, plot_corner, plot_fit, plot_prior_predictive


def parse_args():
    """
    Parse the commandline arguments.

    Returns
    -------
    args: populated namespace
        The commandline arguments.
    """

    parser = argparse.ArgumentParser(
        description="Fit distribution data.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "filename",
        type=str,
        help="Name of file to process. The input file must be produced by the fluence time series option of plot-profilestack.",
    )

    parser.add_argument(
        "--fast",
        dest="fast",
        action="store_true",
        default=False,
        help="Enable fast processing. This reduces the number of MCMC steps drastically.",
    )

    parser.add_argument(
        "--labels",
        dest="labels",
        type=str,
        nargs="+",
        metavar=("name"),
        default=None,
        help="The labels to use for each input file.",
    )

    parser.add_argument(
        "--mean",
        dest="mean",
        type=float,
        metavar=("value"),
        default=1.0,
        help="The global mean fluence to divide the histograms by.",
    )

    parser.add_argument(
        "--meanthresh",
        dest="mean_thresh",
        type=float,
        metavar=("value"),
        default=-3.0,
        help="Ignore fluence data below this mean fluence threshold, i.e. select only data where fluence / mean > meanthresh.",
    )

    parser.add_argument(
        "--model",
        dest="model",
        choices=["NL", "NN", "NNL"],
        default="NNL",
        help="Use the specified distribution model, where N denotes a Normal and L a Lognormal component. For instance, the default NNL model consists of two Normal and one Lognormal distributions.",
    )

    # options that affect the output formatting
    output = parser.add_argument_group(title="Output formatting")

    output.add_argument(
        "--ccdf",
        dest="ccdf",
        action="store_true",
        default=False,
        help="Show the CCDF (cumulative counts) instead of the PDF (differential counts).",
    )

    output.add_argument(
        "--log",
        dest="log",
        action="store_true",
        default=False,
        help="Show histograms in double logarithmic scale.",
    )

    output.add_argument(
        "--nbin",
        dest="nbin",
        type=int,
        metavar=("value"),
        default=50,
        help="The number of histogram bins to use.",
    )

    output.add_argument(
        "-o",
        "--output",
        dest="output",
        action="store_true",
        default=False,
        help="Output plots to file rather than to screen.",
    )

    parser.add_argument(
        "--title",
        dest="title",
        type=str,
        metavar=("text"),
        default=None,
        help="Set a custom figure title.",
    )

    args = parser.parse_args()

    return args


def check_args(args):
    """
    Sanity check the commandline arguments.

    Parameters
    ----------
    args: populated namespace
        The commandline arguments.
    """

    log = logging.getLogger("fitpdf.fit_pdf")

    # check the labels
    if args.labels is not None:
        if len(args.labels) == len(args.files):
            pass
        else:
            log.error(
                "The number of labels is invalid: {0}, {1}".format(
                    len(args.files), len(args.labels)
                )
            )
            sys.exit(1)

    # check the mean
    if args.mean > 0:
        pass
    else:
        log.error(f"The mean fluence is invalid: {args.mean}")
        sys.exit(1)

    # check that file exist
    if not os.path.isfile(args.filename):
        log.error(f"File does not exist: {args.filename}")
        sys.exit(1)


def fit_pe_dist(t_data, t_offp, params):
    """
    Fit pulse-energy distribution.

    Parameters
    ----------
    t_data: ~np.array of float
        The input data.
    t_offp: ~np.array of float
        The off-pulse data.
    params: dict
        Additional parameters that influence the processing.
    """

    data = t_data.copy()
    offp = t_offp.copy()

    # model selection
    if params["model"] == "NL":
        mobj = fmodels.NL()
    elif params["model"] == "NN":
        mobj = fmodels.NN()
    elif params["model"] == "NNL":
        mobj = fmodels.NNL()
    else:
        raise NotImplementedError("Model not implemented: %s", params["model"])

    model = mobj.get_model(data, offp)

    print(f"All RVs: {model.basic_RVs}")
    print(f"Free RVs: {model.free_RVs}")
    print(f"Observed RVs: {model.observed_RVs}")
    print(f"Initial point: {model.initial_point()}")

    config = {}

    if params["fast"]:
        config["draws"] = 700
        config["tune"] = 700
    else:
        config["draws"] = 10000
        config["tune"] = 2000

    with model:
        idata = pm.sample(
            draws=config["draws"],
            tune=config["tune"],
            chains=4,
            init="advi+adapt_diag",
            nuts={"target_accept": 0.9},
        )
        pm.compute_log_likelihood(idata)

    _df_result = az.summary(idata)
    print(_df_result)
    _df_result.to_csv("fit_result.csv")

    plot_chains(idata, params)
    plot_corner(idata, params)

    # compute prior predictive samples
    with model:
        # sample all the parameters
        pp = pm.sample_prior_predictive()

    assert hasattr(pp, "prior_predictive")

    plot_prior_predictive(pp, data, offp, params)

    # compute posterior predictive samples
    thinned_idata = idata.sel(draw=slice(None, None, 20))

    with model:
        pp = pm.sample_posterior_predictive(thinned_idata, var_names=["obs"])
        idata.extend(pp)

    assert hasattr(idata, "posterior_predictive")

    # save idata to file
    _filename = "idata_{0}.nc".format(params["model"])
    az.to_netcdf(idata, _filename)

    plot_fit(mobj, idata, offp, params)

    # output the fit parameters
    print("\nFit parameters")
    for icomp in range(mobj.ncomp):
        # weight
        _samples = idata.posterior["w"]
        quantiles = _samples.sel(component=icomp).quantile(
            q=[0.16, 0.5, 0.84], dim=("chain", "draw")
        )
        error = np.maximum(
            np.abs(quantiles[1] - quantiles[0]), np.abs(quantiles[2] - quantiles[1])
        )
        weight = {"value": quantiles[1], "error": error}

        # mode
        _samples = mobj.get_mode(idata.posterior["mu"], idata.posterior["sigma"], icomp)
        quantiles = _samples.sel(component=icomp).quantile(
            q=[0.16, 0.5, 0.84], dim=("chain", "draw")
        )
        error = np.maximum(
            np.abs(quantiles[1] - quantiles[0]), np.abs(quantiles[2] - quantiles[1])
        )
        mode = {"value": quantiles[1], "error": error}

        print(
            "Component {0}: {1:.3f} +- {2:.3f}, {3:.3f} +- {4:.3f}".format(
                icomp, weight["value"], weight["error"], mode["value"], mode["error"]
            )
        )


#
# MAIN
#


def main():
    # start signal handler
    signal.signal(signal.SIGINT, signal_handler)

    # set up logging
    configure_logging()
    log = logging.getLogger("fitpdf.fit_pdf")

    # handle command line arguments
    args = parse_args()

    # sanity check command line arguments
    check_args(args)

    # tweak the matplotlib output formatting
    customise_matplotlib_format()

    params = {
        "ccdf": args.ccdf,
        "dpi": 300,
        "fast": args.fast,
        "labels": args.labels,
        "log": args.log,
        "mean": args.mean,
        "mean_thresh": args.mean_thresh,
        "model": args.model,
        "nbin": args.nbin,
        "output": args.output,
        "publish": False,
        "title": args.title,
    }

    print(f"Processing: {args.filename}")
    df = pd.read_csv(args.filename)
    df["filename"] = args.filename

    _data, _offp = plot_pe_dist([df], params)

    fit_pe_dist(_data / params["mean"], _offp / params["mean"], params)

    plt.show()

    log.info("All done.")


if __name__ == "__main__":
    main()



#
#   2025 Fabian Jankowski
#   Statistics related helper functions.
#

import numpy as np


def get_adaptive_bandwidth(t_data, min_bw):
    """
    Compute an adaptive bandwidth for kernel density estimation (KDE).

    The bandwidth changes based on the density of data points.

    Parameters
    ----------
    t_data: ~np.array
        The input data.
    min_bw: float
        The minimum bandwidth to enforce.

    Returns
    -------
    bandwidths: ~np.array
        The kernel bandwidth for each data point.
    """

    data = t_data.copy()
    data = np.sort(data)

    assert min_bw > 0

    bandwidths = np.zeros(len(data))
    bandwidths[:-1] = 0.5 * np.diff(data)
    bandwidths[-1] = bandwidths[-2]
    bandwidths = np.clip(bandwidths, a_min=min_bw, a_max=None)

    return bandwidths



#
#   2025 Fabian Jankowski
#   Distribution model template.
#

import logging


class Model(object):
    name = "model"

    def __init__(self):
        """
        Model distribution.
        """

        self.__log = logging.getLogger("fitpdf.models")

        self.ncomp = None

    def __repr__(self):
        """
        Representation of the object.
        """

        info_dict = {"bla": "XXX"}

        info_str = "{0}".format(info_dict)

        return info_str

    def __str__(self):
        """
        String representation of the object.
        """

        info_str = "{0}: {1}".format(self.name, repr(self))

        return info_str

    def get_model(self, t_data, t_offp):
        """
        Construct a mixture model.

        Parameters
        ----------
        t_data: ~np.array of float
            The input data to be fit.
        t_offp: ~np.array of float
            The off-pulse data.

        Returns
        -------
        model: ~pm.Model
            A mixture model.
        """
        pass

    def get_analytic_pdf(self, x, w, mu, sigma, icomp):
        """
        Get the analytic PDF.

        Returns
        -------
        pdf: ~np.array of float
            The model PDF evaluated at the `x` values.
        """
        pass

    def get_mode(self, mu, sigma, icomp):
        """
        Compute the mode of the model component.

        Returns
        -------
        mode: ~np.array of float
            The mode of the model component.
        """
        pass
